open! Dune_exe
open! Dune_exe__Describe__
let arg =
  Arg.(
    value
    & opt (some string) None
    & info
        [ "lang" ]
        ~docv:"VERSION"
        ~doc:
          (Some
             "This argument has no effect and is deprecated. It exists solely for \
              backwards compatibility."))
;;
