open! Action_plugin__
include struct
  open Dune_engine
  module Action = Action
  module Dep = Dep
  module File_selector = File_selector
  module Rpc = Rpc
  module Process = Process
  module Clflags = Clflags
  module Done_or_more_deps = Done_or_more_deps
  module Exec = Action.Ext.Exec
end

include Stdune
module Glob = Dune_glob.V1
