open! Dune_lang__
include Stdune

include struct
  open Dune_sexp
  module Decoder = Decoder
  module Encoder = Encoder
  module Syntax = Syntax
  module Conv = Conv
  module Cst = Cst
  module Parser = Parser
  module Ast = Ast
  module Template = Template
end

include struct
  module Mode = Ocaml.Mode
end
