open! Dune_rules__
let store_orig_src_dir = ref false
let ignore_promoted_rules = ref false
let promote_install_files = ref false
let display = Dune_engine.Clflags.display
let capture_outputs = Dune_engine.Clflags.capture_outputs
let debug_package_logs = ref false
let concurrency = ref 1
