open! Dune_rpc_impl__
include Stdune
module Console = Console
module Dune_rpc = Dune_rpc.Private
include Dune_scheduler

include struct
  open Dune_engine
  module Build_system = Build_system
  module Build_system_error = Build_system_error
  module Running_jobs = Running_jobs
end

include struct
  open Rpc
  module Where = Where
  module Client = Rpc.Client
end
