open! Dune_lang__
include Dune_sexp
module Lib_name = Lib_name
module Alias = Alias
module Format = Format
module Stanza = Stanza
module Glob = Glob
module String_with_vars = String_with_vars
module Pform = Pform
module Action = Action
module Dune_file_script = Dune_file_script
module Path = Path
module Value = Value
module Blang = Blang
module Binary_kind = Binary_kind
module Package_constraint = Package_constraint
module Package_name = Package_name
module Package_dependency = Package_dependency
module Package_version = Package_version
module Package_variable_name = Package_variable_name
module Lib_kind = Lib_kind
module Lib_dep = Lib_dep
module Pkg = Pkg
module Ordered_set_lang = Ordered_set_lang
module Format_config = Format_config
module Bindings = Bindings
module Subst_config = Subst_config
module Locks = Locks
module Profile = Profile
module Slang = Slang
module Targets_spec = Targets_spec
module Wrapped = Wrapped
module Visibility = Visibility
module Dep_conf = Dep_conf
module Relop = Relop
module Cond = Cond
module Toggle = Toggle
module Site = Site
module Warning = Warning
module Source_kind = Source_kind
module Package_info = Package_info
module Section = Section
module Package = Package
module Dune_project_name = Dune_project_name
module Dialect = Dialect
module Lib_mode = Lib_mode
module Melange = Melange
module Module_name = Module_name
module Pin_stanza = Pin_stanza
module Preprocess = Preprocess
module Link_flags = Link_flags
module Dune_project = Dune_project
module File_binding = File_binding
module Foreign_language = Foreign_language
module Ocaml_flags = Ocaml_flags
module Coq_env = Coq_env
module Rocq_env = Rocq_env
module Menhir_env = Menhir_env
module Dune_env = Dune_env
module Js_of_ocaml = Js_of_ocaml
module Menhir = Menhir
module Rule_mode = Rule_mode
module Rule_mode_decoder = Rule_mode_decoder
module Mode_conf = Mode_conf
module Oxcaml = Oxcaml
module Modules_settings = Modules_settings
module Stanza_pkg = Stanza_pkg
module Package_mask = Package_mask
module Unreleased = Unreleased

(* CR-someday rgrinberg: perhaps wrap these under [Stanzas]? *)
module Copy_files = Copy_files
module Enabled_if = Enabled_if
module Alias_conf = Alias_conf
module Include_subdirs = Include_subdirs
